program rijmen;                                                  { version 1.2 }


(******************************************************** 1992 J.C. Kessels ****
Rhyme editor.


J.C. Kessels
Philips de Goedelaan 7
5615 PN  Eindhoven
Netherlands
usenet: jeroenk@cnps.cnps.philips.nl
*******************************************************************************)



uses dos;



(*******************************************************************************
** Language dependent settings (Dutch).                                       **
*******************************************************************************)

const { Dutch }
  Vowels        : set of char = ['a','e','i','o','u','y'];
  StrCopyright  = 'RIJMEN v1.2, 1992 J.C. Kessels';         { Name of program. }
  StrLine       = 'regel ';                                { Line number text. }
  StrStart      = 'U kunt nu tekst intikken....';     { You may start rhyming. }
  { Busy printing.}
  StrPrinting   = 'Druk op een toets om te stoppen met printen....';
  StrEof        = 'U heeft het maximum van 500 regels bereikt.';  { Last line. }
  StrReserved   = 'Niet toegestaan: de file eindigt op ''.dut'', ''.pas'', ''.exe'', of ''.com''.';
  StrAccess     = 'Ongeldige filenaam: toegang is door DOS geweigerd.';
  StrClose      = 'Dank U voor het gebruiken van: RIJMEN!';
  HintFileExt   = '.dut';               { Extension of name of words database. }
  HelpFileName  = 'lees.mij';

  { The help texts shown when F1 is pressed, and the permanent help text shown
    at the bottom line of the screen. }
  HelpLines1    = 7;
  Help1 : array[0..HelpLines1] of string[80] =
    ('F1=help ESC=verlaat RIJMEN',
     'Pijltjes:  Beweeg heen en weer.              Home:  Naar begin van de regel.',
     'Enter:     Splits de regel.                  End:   Naar einde van de regel.',
     'Backspace: Verwijder achterwaards.           PgUp:  Pagina omhoog.',
     'Del:       Verwijder voorwaards.             PgDn:  Pagina omlaag.',
     '',
     'F7:        Print het rijmpje.',
     'Esc:       Bewaar het rijmpje op disk en verlaat het RIJMEN programma.');
  HelpLines2    = 6;
  Help2 : array[0..HelpLines2] of string[76] =
    ('F1=help  ENTER=ga rijmen  ESC=verlaat RIJMEN',
     'Enter:     Stop met het lezen van de documentatie en begin te rijmen.',
     'Pijltjes:  Beweeg op en neer.',
     'PgUp:      Pagina omhoog.',
     'PgDn:      Pagina omlaag.',
     'F7:        Print de documentatie.',
     'Esc:       Verlaat het RIJMEN programma.');
  HelpLines3    = 5;
  Help3 : array[0..HelpLines3] of string[76] =
    ('F1=help  ENTER=accepteer  ESC=verlaat RIJMEN',
     'Enter:     Als U klaar bent met het ingeven van de filenaam.',
     'Pijltjes:  Beweeg links en rechts.',
     'Backspace: Verwijder achterwaards.',
     'Del:       Verwijder voorwaards.',
     'ESC:       Verlaat het RIJMEN programma.');

  { The messages shown when asking for a filename. }
  AskFilenameLines = 8;
  AskFilename : array[1..AskFilenameLines] of string[76] = (
    'Enter de filenaam van een bestaand of nieuw rijm:',
    '',
    '',
    '',
    'De filenaam moet aan de volgende voorwaarden voldoen:',
    ' - Het moet een geldige DOS filenaam zijn, eventueel inclusief drive en pad.',
    ' - De file moet leesbaar en schrijfbaar zijn.',
    ' - Het mag geen ''.dut'', ''.pas'', ''.exe'', of ''.com'' file zijn.');






(*******************************************************************************
** Constants and global variables.                                            **
*******************************************************************************)

const
  RhymeLinesMax = 500;                     { The maximum number of text lines. }
  RhymeCharsMax = 76;             { The maximum number of characters per line. }


type
  Buf = array[0..65534] of char;                        { General buffer type. }


var
  Rhyme        : array[1..RhymeLinesMax] of string[RhymeCharsMax]; { The text. }
  RhymeTop     : word;               { The top of the page (1..RhymeLinesMax). }
  RhymeLine    : word;                   { The cursor line (1..RhymeLinesMax). }
  RhymeCursor  : word;     { The column into the RhymeLine (1..RhymeCharsMax). }

  HintWord     : string;                        { The word we are looking for. }
  HintWordBase : string;              { The minimum piece of the word we want. }
  HintWordBaseGroups : word;           { The number of groups in HintWordBase. }
  HintBusy     : boolean;                  { TRUE: busy searching rhyme words. }
  HintFile     : file of char;                   { The file with the wordlist. }
  HintFileName : string;                  { The name of the current hint file. }
  HintBase     : longint;                         { Offset into wordlist file. }
  HintInterval : longint;                { Search interval into wordlist file. }
  HintMessage  : string;                      { Message in the separator line. }
  HintLines    : array[1..50] of string[RhymeCharsMax];         { Rhyme words. }
  HintLinesMax : word;                 { The number of lines with rhyme words. }
  HintLinesRefresh : boolean;                     { TRUE: There are new words. }
  HintWords    : array[1..50] of record
    ss                : string[40];
    AlikeGroups       : word;
    LastEqualGroup    : word;
    UnequalCharacters : word;
    end;
  HintWordsTop : word;

  VidStore     : ^Buf;                                 { Video screen storage. }
  OldCursor    : word;                                   { Saved cursor shape. }
  xy           : word;                                 { Cursor to the screen. }
  colour       : byte;                     { Current colour for screen writes. }
  vidseg       : word;                            { Segment of the screen RAM. }
  LPS          : word;                               { Title lines per screen. }
  CPL          : word;                           { Characters per screen line. }

  regs         : registers;                    { Used for DOS and BIOS access. }
  HelpMessage  : string[76];                       { Bottom line help message. }
  ErrorMessage : string[76];                                  { Error message. }
  HintFilePath : string;                                { Path to the lexicon. }







(*******************************************************************************
** Basic subroutines.                                                         **
*******************************************************************************)

function KeyPressed : boolean;
{ Return TRUE if a key is waiting in the keyboard buffer. }
begin
regs.ah := 1;
intr($16,regs);
KeyPressed := (regs.flags and fzero) = 0;
end;





function Left(s : string; width : byte) : string;
{ Return Width characters from input string. Add trailing spaces if necessary.
  }
begin
if width > length(s) then Fillchar(s[length(s)+1],width-length(s),32);
s[0] := chr(width);
Left := s;
end;






function Lower(s : string) : string;
{ Translate string to lowercase. }
var
  i : word;
begin
for i := 1 to length(s) do
  if s[i] IN ['A'..'Z'] then s[i] := chr(ord(s[i])+32);
Lower := s;
end;




function StripT(s : string) : string;
{ Strip input string of trailing spaces. }
var
  i : word;
begin
i := length(s) + 1;
repeat dec(i) until (i < 1) or (s[i] <> ' ');
if i < length(s) then delete(s,i+1,length(s)-i);
StripT := s;
end;




function Reverse(InputString : string): string;
{ Reverse the characters in the InputString, so that the last character becomes
  the first and vice versa. }
var
  i, j : word;
  OutputString : string;
begin
j := length(InputString);
for i := 1 to j do OutputString[j-i+1] := InputString[i];
OutputString[0] := chr(j);
Reverse := OutputString;
end;




function GroupsCount(ss : string) : word;
{ Return the number of vowel and consonant groups in the input word. The input
  word must be in all lowercase. }
var
  groups, i : word;
  VowelGroup : boolean;
begin
groups := 1;
if ss[1] in Vowels
  then VowelGroup := true
  else VowelGroup := false;
for i := 1 to length(ss) do
  begin
  if ss[i] in Vowels
    then
      begin
      if not VowelGroup then
        begin
        VowelGroup := not VowelGroup;
        inc(groups);
        end;
      end
    else
      begin
      if VowelGroup then
        begin
        VowelGroup := not VowelGroup;
        inc(groups);
        end;
      end;
  end;
GroupsCount := groups;
end;




function RightGroups(ss : string; count : word) : string;
{ Return the substring from the right of the input string that contains "count"
  vowels-groups and consonants-groups. The input string must be all lower-case.
  }
var
  i : word;
  VowelGroup : boolean;
begin
i := length(ss);
if ss[i] in Vowels
  then VowelGroup := true
  else VowelGroup := false;
while (count > 0) and (i > 0) do
  begin
  if ss[i] in Vowels
    then
      begin
      if not VowelGroup then
        begin
        VowelGroup := not VowelGroup;
        dec(count);
        end;
      end
    else
      begin
      if VowelGroup then
        begin
        VowelGroup := not VowelGroup;
        dec(count);
        end;
      end;
  if count > 0 then dec(i);
  end;
RightGroups := copy(ss,i+1,255);
end;




procedure Beep;               
{ The phone is ringing. }
var
  ToneCount      : word;                        { The number of tones to play. }
  t              : longint;                     { Temporary calculus variable. }
  Hz             : word;                         { The frequency of the sound. }
  OneMillisecond : word;        { The number of loops needed for one clocktic. }
  count          : word;                        { Delay loop internal counter. }
  OldDosTimer    : byte;                         { Remembered dos timer value. }
  DosTimer       : byte absolute $0040:$006C;           { The DOS timer (LSB). }
begin
OldDosTimer := DosTimer;          { First wait for the DOS-timer to tick over. }
repeat until DosTimer <> OldDosTimer;
OldDosTimer := DosTimer;                       { Load Delay-loop variables.... }
count := 1;
{ and run the actual Delay-loop: }
while (DosTimer = OldDosTimer) and (count > 0) do inc(count);
OneMillisecond := count;                   { Save the Delay-loop tuning value. }

Hz := 1200;                                                { Start at 1200 Hz. }
for ToneCount := 1 to 10 do
  begin
  t := 1193180 div Hz;                   { Timer chip setting := 1193180 / Hz. }
  count := t;
  port[$43] := 182;
  port[$42] := lo(count);
  port[$42] := hi(count);
  port[$61] := port[$61] or 3;                         { Activate loudspeaker. }

  t := OneMillisecond;                           { Calculate outer loop count. }
  t := t * 40 * 18 div 1000;                      { Delay for 40 milliseconds. }
  while t > 0 do                                                 { Outer loop. }
    begin
    if t > 60000                                 { Determine inner loop count. }
      then count := 60000
      else count := t;
    t := t - count;
    OldDosTimer := mem[$0000:$0000];                             { Inner loop. }
    while (mem[$0000:$0000] = OldDosTimer) and (count > 0) do dec(count);
    end;

  if Hz = 1200                              { Toggle between 1000 and 1200 Hz. }
    then Hz := 1000
    else Hz := 1200;
  end;

port[$61] := port[$61] and $FC;                    { Turn off the loudspeaker. }
end;






(*******************************************************************************
** The rhyme-search engine.                                                   **
*******************************************************************************)



function StartHintSearch(ss : string) : boolean;
{ Determine the word to be found. If equal to the last word to be found, than
  continue that search. Otherwise start a new search. }
var
  s1 : string;
  i, j : word;
begin

s1 := '';             { Extract the word to be found from the end of the line. }
if ss <> '' then
  begin
  i := length(ss);
  j := 0;
  { Backskip non-characters. }
  while (i > 0) and (not (ss[i] in ['a'..'z','A'..'Z'])) do dec(i);
  { Count number of characters. }
  while (i > 0) and (ss[i] in ['a'..'z','A'..'Z']) do
    begin
    dec(i);
    inc(j);
    end;
  if j > 0 then s1 := Lower(copy(ss,i+1,j));            { Copy the rhyme word. }
  end;

StartHintSearch := false;
if s1 = HintWord then exit;    { Exit if we are already looking for this word. }

{ Start a new search. }
if HintBusy then close(HintFile);                  { Close the words database. }
if s1 <> ''
  then HintBusy := true
  else HintBusy := false;
HintWord        := s1;
HintWordBaseGroups := 2;
HintWordBase    := RightGroups(s1,2);
if HintWordBase = s1 then
  begin
  HintWordBase := RightGroups(s1,1);
  HintWordBaseGroups := 1;
  end;
str(GroupsCount(HintWord),HintFileName);
HintFileName    := Fexpand(HintFilePath + 'w' + HintFileName + HintFileExt);
HintMessage     := s1;
HintLinesMax    := 0;
HintLinesRefresh := true;
for i := 1 to 50 do HintWords[i].ss := '';
HintWordsTop := 0;
if HintBusy then                                    { Open the words database. }
  begin
  {$I-}
  assign(HintFile,HintFileName);
  reset(HintFile);
  if IOresult <> 0
    then HintBusy := false
    else
      begin
      HintBase      := filesize(HintFile) shr 1;
      HintInterval  := HintBase shr 1;
      end;
  {$I-}
  end;

StartHintSearch := true;                                               { Exit. }
end;





procedure StopHintSearch(message: string);
{ Stop the search for hint words. Set a new message text. }
begin
HintLinesMax := 0;                                    { Delete all hint lines. }
HintLinesRefresh := true;                              { Refresh display flag. }
if HintBusy then close(HintFile);                  { Close the words database. }
HintBusy    := false;                              { Not searching for a hint. }
HintWord    := '';                                    { The hintword is empty. }
HintMessage := message;                                     { Set new message. }
end;




procedure HintSearchZoom;
{ Iterate the search for the HintWord in the words database. This is done by
  starting in the middle of the file, and moving up or down in continually
  smaller steps until the first-smaller or equal entry is found. The "HintBase"
  variable contains the location we are looking at, the "HintInterval" variable
  the step size with which to move up or down. }
var
  ch : char;
  ss : string;
  i, j : word;
begin
seek(HintFile,HintBase);                                 { Seek into the file. }
repeat read(HintFile,ch); until ch = #10;                  { Walk up to CR/LF. }

ss := '';                                              { Fetch word from file. }
repeat
  read(HintFile,ch);
  if ch <> #13 then ss := ss + ch;
until ch = #13;

{ If the word is greater-equal than what we are looking for, then move
  backwards in the file, else move forwards. }
i := length(ss);
j := length(HintWordBase);
while (i > 0) and (j > 0) and (ss[i] = HintWordBase[j]) do
  begin
  dec(i);
  dec(j);
  end;
if (j = 0) or ((i <> 0) and (ss[i] > HintWordBase[j]))
  then
    begin                                                    { Move backwards. }
    if HintBase > HintInterval
      then HintBase := HintBase - HintInterval
      else HintBase := 0;
    HintInterval := HintInterval shr 1;
    end
  else
    begin                                                     { Move forwards. }
    HintBase := HintBase + HintInterval;
    HintInterval := HintInterval shr 1;
    end;

{ We are done if the search interval has decreased to 25 characters. We could
  keep searching until the interval has decreased to 1 character, but precision
  like that isn't necessary. }
if HintInterval < 25 then
  begin
  HintInterval := 0;
  if HintBase > 50
    then HintBase := HintBase - 50
    else HintBase := 0;
  seek(HintFile,HintBase);
  if HintBase > 0 then repeat read(HintFile,ch); until ch = #10;
  end;
end;




procedure AnalyzeHintWord(ss : string);
{ Add a word to the HintWords array of already found (rhyming) words. The array
  is sorted by rhyme, the most rhyming words are the first words in the array.
  }
{ How much a word "rhymes" is determined by counting a number of "differences"
  between the two words. Counting is based on (character-)groups. A group is a
  sequence of vowels or a sequence of consonants. The following differences are
  counted, listed in importance for rhyming: }
{ 1. The total number of groups that are equal between two words. Higher means
     better rhyme.
  2. The number of the last equal groups between the two words, counted from
     the end of the words. Higher means better rhyme.
  3. The number of characters in the unequal groups between the two words.
     Less means better rhyme. }
 var
  Group        : word;                          { Number of the current group. }
  AlikeGroups       : word;                      { The number of equal groups. }
  LastEqualGroup    : word;                            { The last-equal group. }
  UnequalCharacters : word;                        { Total unequal characters. }
  GroupIsEqual : boolean;                             { TRUE: groups are equal.}
  VowelGroup   : boolean;                 { TRUE: currently processing Vowels. }
  CharsInGroup : word;                    { Number of characters in the group. }
  av, bv       : boolean;                                    { Work variables. }
  i, j         : word;
  s3 : string;
begin
if ss = HintWord then exit;                   { Exit if equal to the HintWord. }

AlikeGroups := 0;                                            { Clear counters. }
LastEqualGroup := 0;
UnequalCharacters := 0;
GroupIsEqual := true;
Group        := 1;
CharsInGroup := 0;

{ Walk through the HintWord and the (current-)word one character at a time,
  starting at the end of both words. Compare characters as we go, and increment
  the appropriate counters. }
i := length(ss);
j := length(HintWord);
{ Determine if the "current-group" that we start with is a Vowel or a Consonant
  group. }
VowelGroup := HintWord[j] in Vowels;
while (i > 0) or (j > 0) do             { While the words are not exhausted....}
  begin
  { Determine for both words if we are currently pointing to a vowel or a
    consonant. }
  if i > 0 then av := ss[i] in Vowels;
  if j > 0 then bv := Hintword[j] in Vowels;

  if av = VowelGroup                  { If the word is in the current-group....}
    then
      begin              { then if the HintWord is not in the current-group....}
      if (bv <> VowelGroup) or (i = 0) or (j = 0) or (ss[i] <> HintWord[j]) then
        GroupIsEqual := false;                { then the groups are not equal. }
      { If the Hintword is in the current-group then advance it's pointer. }
      if (bv = VowelGroup) and (j > 0) then dec(j);
      { Advance the word pointer, and increment the group-characters
        counter. }
      if i > 0 then
        begin
        inc(CharsInGroup);
        dec(i);
        end;
      end
    else                      { Else (the word is NOT in the current-group)....}
      if bv = VowelGroup          { If the Hintword is in the current-group....}
        then
          begin
          GroupIsEqual := false;              { then the groups are not equal. }
          if j > 0 then dec(j);         { Advance the pointer of the HintWord. }
          end
        else               { Else (both words are NOT in the current group)....}
          begin
          VowelGroup := not VowelGroup;              { Toggle "current-group". }
          { If all the characters in the group were equal between the two
            words, then increment the "AlikeGroups" counter. If not, then
            increment the "UnEqualCharacters" counter with the number of
            characters in the group, and if not set already then set the
            "LastEqualGroup" counter to the current group minus one. }
          if GroupIsEqual
            then inc(AlikeGroups)
            else
              begin
              UnEqualCharacters := UnequalCharacters + CharsInGroup;
              if LastEqualGroup = 0 then LastEqualGroup := group - 1;
              end;
          GroupIsEqual := true;                             { Reset variables. }
          CharsInGroup := 0;
          inc(group);                               { Increment group counter. }
          if (i = 0) or (j = 0) or (ss[i] <> HintWord[j]) then
            GroupIsEqual := false;
          if j > 0 then dec(j);                    { Advance Hintword pointer. }
          { Advance word pointer, and increment the group-characters counter. }
          if i > 0 then
            begin
            inc(CharsInGroup);
            dec(i);
            end;
          end;
  end;
{ Finish the comparison now that we have reached the end of one of the two
  words. }
if GroupIsEqual
  then inc(AlikeGroups)
  else UnEqualCharacters := UnequalCharacters + CharsInGroup;
if LastEqualGroup = 0 then LastEqualGroup := group - 1;

{ Search the proper place to insert this word in the HintWords array. If all
  else fails, then sort alphabetically. }
i := HintwordsTop;
while (i > 0) and (
   (Hintwords[i].AlikeGroups < AlikeGroups) or
   ((Hintwords[i].AlikeGroups = AlikeGroups) and
    (Hintwords[i].LastEqualGroup < LastEqualGroup)) or
   ((Hintwords[i].AlikeGroups = AlikeGroups) and
    (Hintwords[i].LastEqualGroup = LastEqualGroup) and
    (HintWords[i].UnequalCharacters > UnequalCharacters)) or
   ((Hintwords[i].AlikeGroups = AlikeGroups) and
    (Hintwords[i].LastEqualGroup = LastEqualGroup) and
    (HintWords[i].UnequalCharacters = UnequalCharacters) and
    (HintWords[i].ss > ss))) do dec(i);

{ Exit if the last word in the HintWords array rhymes more than this word. }
if i = 50 then exit;

{ Insert the word into the HintWords array. Display-update happens somewhere
  else. }
HintLinesRefresh := true;
for j := 49 downto i+1 do HintWords[j+1] := HintWords[j];
inc(i);
HintWords[i].ss := ss;
HintWords[i].AlikeGroups := AlikeGroups;
HintWords[i].LastEqualGroup := LastEqualGroup;
HintWords[i].UnequalCharacters := UnequalCharacters;
if HintwordsTop < 50 then inc(HintwordsTop);

end;





procedure FillHintLines;
{ Fill the hintlines with the words found so far. This subroutine is called
  just before the hintlines are displayed. }
var
  i : word;
begin
HintLinesMax := 0;
for i := 1 to HintWordsTop do
  begin
  if (HintLinesMax = 0) or (length(HintLines[HintLinesMax]) + 1 +
      length(HintWords[i].ss) > RhymeCharsMax) then
    begin
    inc(HintLinesMax);
    HintLines[HintLinesMax] := '';
    end;
  if length(HintLines[HintLinesMax]) = 0
    then HintLines[HintLinesMax] := HintWords[i].ss
    else HintLines[HintLinesMax] := HintLines[HintLinesMax] + ' ' +
         HintWords[i].ss;
  end;
end;




procedure HintSearchWord;
{ Walk through the datafile testing the words for similarity with the word we
  are looking for. Stop searching at the first larger word in the file. }
var
  ss : string;
  ch : char;
  i, j : word;
begin
ss := '';                                            { Fetch word from file. }
i := 0;
repeat
  read(HintFile,ch);
  if ch <> #13 then
    begin
    inc(i);
    ss[i] := ch;
    end;
until (ch = #13) or eof(HintFile);
ss[0] := chr(i);
read(HintFile,ch);

{ Test if the incoming word has the same ending as the word we are looking for.
  If so, then further test the word for it's rhyming properties. }
if HintWordBase = RightGroups(ss,HintWordBaseGroups) then
  begin
  AnalyzeHintWord(ss);                                 { Add word to wordlist. }
  if eof(HintFile) then       { If at the end of the file then stop searching. }
    begin
    close(HintFile);                               { Close the words database. }
    HintBusy := false;
    end;
  exit;
  end;

{ Test if we have exhausted the word database for rhyme words. Since the words
  database is sorted by wordending, we only have to test if the current word is
  larger than the base of the word we are searching for. }
i := length(HintWordBase);
j := length(ss);
while (i > 0) and (j > 0) and (HintWordBase[i] = ss[j]) do
  begin
  dec(i);
  dec(j);
  end;
if ((i > 0) and (j > 0) and (HintWordBase[i] < ss[j])) or
   eof(HintFile) then
  begin
  close(HintFile);                                 { Close the words database. }
  HintBusy := false;
  end;
end;






(*******************************************************************************
** Video subroutines.                                                         **
*******************************************************************************)

procedure Show(s : string);
{ Display string "s" at "xy", using "colour". This routine uses DMA into the
  video memory. }
begin
Inline(
  $8E/$06/>VIDSEG/       {mov  es,[>vidseg]   ; Fetch video segment in ES.}
  $8B/$3E/>XY/           {mov  di,[>xy]       ; Fetch video offset in DI.}
  $8A/$26/>COLOUR/       {mov  ah,[>colour]   ; Fetch video colour in AH.}
  $1E/                   {push ds             ; Setup DS to stack segment.}
  $8C/$D1/               {mov  cx,ss}
  $8E/$D9/               {mov  ds,cx}
  $8A/$8E/>S/            {mov  cl,[bp+>s]     ; Fetch string size in CX.}
  $30/$ED/               {xor  ch,ch}
  $8D/$B6/>S+1/          {lea  si,[bp+>s+1]   ; Fetch string address in SI.}
  $E3/$04/               {jcxz l2             ; Skip if zero length.}
                         {l1:}
  $AC/                   {lodsb               ; Fetch character from string.}
  $AB/                   {stosw               ; Show character.}
  $E2/$FC/               {loop l1             ; Next character.}
                         {l2:}
  $1F/                   {pop  ds             ; Restore DS.}
  $89/$3E/>XY);          {mov  [>xy],di       ; Store new XY.}
end;





function IsTrident : boolean;
{ Return TRUE if the card is a Trident SVGA card. We do this by testing if bit
  1 in the Mode Control #1 register is inverting. If inverting, then it is a
  Trident. }
var
  OldValue : byte;
  Flag : boolean;
begin
port[$3C4] := $0E;                          { Select Mode Control #1 register. }
OldValue   := port[$3C5];                           { Save it's current value. }
port[$3C5] := 0;                                         { Reset it to zero's. }
Flag       := ((port[$3C5] and $0F) = $02);    { Read back & test if inverted. }
if Flag then OldValue := OldValue xor $02;         { If Trident then flip bit. }
port[$3C5] := OldValue;                     { Reset it to it's original value. }
IsTrident := Flag;                                     { Return TRUE or FALSE. }
end;




procedure InitVideo;
{ Initialise the video. If not 80x25 then switch to it. Store the screen.
  Hide the cursor. }
var
  VidMode : byte;
begin
regs.ah := $0F;                                    { Fetch current video mode. }
intr($10,regs);
VidMode := regs.al and $7F;
regs.ah := $03;                { Save current cursor shape. BH is active page. }
intr($10,regs);
OldCursor := regs.cx;
{ If this is one of the special Trident SVGA text modes, then pretend we are in
  video mode 3. }
if (VidMode in [$50..$5A]) and IsTrident then VidMode := 3;
{ If the current video mode is not 0..3 or 7, then switch to video mode 3. }
if not VidMode in [$00..$03,$07] then
  begin
  regs.al := 3;                                      { Switch to video mode 3. }
  regs.ah := 0;
  intr($10,regs);
  VidMode := 3;
  end;
if VidMode <> 7                                       { Compute video segment. }
  then vidseg := $B800 + (memw[$0040:$004E] shr 4)
  else vidseg := $B000 + (memw[$0040:$004E] shr 4);
regs.ah := $0F;                                   { Store characters per line. }
intr($10,regs);
CPL := regs.ah;
LPS := mem[$0040:$0084] + 1;                         { Store lines per screen. }
if (LPS < 25) or (LPS > 60) then LPS := 25;
{ Store current screen. }
if (VidStore = nil) and (MaxAvail > CPL*LPS*2) then
  getmem(VidStore,CPL*LPS*2);
if VidStore <> nil then
  move(mem[vidseg:0],VidStore^,CPL*LPS*2);
regs.cx := $2000;                                               { Hide cursor. }
regs.ah := 1;
intr($10,regs);
colour := $1E;                                              { Reset attribute. }
xy := 0;                                                       { Reset cursor. }
end;




procedure ResetVideo;
{ Reset the video back to it's original contents. Show the cursor. }
begin
if VidStore <> nil then               { Restore the original display contents. }
  move(VidStore^,mem[vidseg:0],CPL*LPS*2);

regs.cx := OldCursor;                           { Reset original cursor chape. }
regs.ah := 1;
intr($10,regs);
end;




procedure ShowCursor(flag : boolean);
{ Turn cursor bar on (TRUE) or off (FALSE). }
begin
xy := (((RhymeLine-RhymeTop+1) * CPL) + RhymeCursor + 1) * 2;
if flag
  then colour := $70
  else colour := $1E;
if RhymeCursor <= length(Rhyme[RhymeLine])
  then Show(Rhyme[RhymeLine,RhymeCursor])
  else Show(' ');
end;




procedure ShowHintlines;
{ Display all the hint lines. }
var
  i   : word;                                                 { Work variable. }
  s   : string;                                               { Work variable. }
begin
xy := (LPS - 2 - HintLinesMax) * CPL * 2;

colour := $13;             { Display the separator line, with the HintMessage. }
Show('');
if HintMessage <> ''
  then
    begin
    Show('');
    if length(HintMessage) > CPL - 6
      then i := CPL - 6
      else i := length(HintMessage);
    if not HintBusy then colour := $1E;
    Show(copy(HintMessage,1,i));
    colour := $13;
    Show('');
    s[0] := chr(CPL - 6 - i);
    end
  else s[0] := chr(CPL - 4);
Fillchar(s[1],ord(s[0]),'');
Show(s+'Ķ');

for i := 1 to HintLinesMax do                         { Display the hintlines. }
  begin
  colour := $13;
  Show(' ');
  if HintBusy
    then colour := $13
    else colour := $1E;
  Show(left(HintLines[i],CPL-4));
  colour := $13;
  Show(' ');
  end;
end;




procedure ShowLinenumber;
{ Show the current line number in the top line. }
var
  s : string;
begin
str(RhymeLine,s);
s := s + '';
xy := 6 + length(StrLine) shl 1;
colour := $13;
Show(s);
end;




procedure ShowLastLine;
var
  i : word;
  s : string;
begin
xy := (LPS - 1) * CPL * 2;
colour := $13;
Show('͵');
{ colour := $1B; }
colour := $13;
s := HelpMessage;
if ErrorMessage <> '' then s := ErrorMessage;
if length(s) > CPL - 6
  then i := CPL - 6
  else i := length(s);
Show(copy(s,1,i));
colour := $13;
s[0] := chr(CPL - 5 - i);
Fillchar(s[1],ord(s[0]),'');
Show(''+s+'');
end;





procedure ShowErrorMessage(Message : string; DoBeep : boolean);
var
  i : word;
  s : string;
begin
if ErrorMessage = Message then exit;
ErrorMessage := Message;

if Message <> '' then
  begin
  xy := (LPS - 1) * CPL * 2 + 6;
  colour := $9B;
  if length(Message) > CPL - 6
    then i := CPL - 6
    else i := length(Message);
  Show(copy(Message,1,i));
  colour := $13;
  s[0] := chr(CPL - 5 - i);
  Fillchar(s[1],ord(s[0]),'');
  Show(''+s+'');
  if DoBeep then beep;
  end;

ShowLastLine;
end;




procedure ShowMenu;
{ Display the screen, with borders. }
var
  i   : word;                                                 { Work variable. }
  s   : string;                                               { Work variable. }
  pagetop : word;                          { Top of the page in the Dir array. }
begin
{ If the cursor is outside the page, then adjust the top of the page. }
if RhymeLine >= RhymeTop + LPS - 3 - HintLinesMax - 3 then
  begin
  RhymeTop := RhymeLine + 3 + HintLinesMax + 3 - LPS;
  if RhymeTop > RhymeLinesMax + 3 + HintLinesMax - LPS then
    RhymeTop := RhymeLinesMax + 3 + HintLinesMax - LPS + 1;
  end;

xy := 0;                                                         { First line. }
colour := $13;
Show('͵'+StrLine);
inc(xy,8);
if length(StrCopyright) > CPL - 12 - length(StrLine)
  then i := CPL - 12 - length(StrLine)
  else i := length(StrCopyright);
s[0] := chr(CPL - 11 - length(StrLine) - i);
Fillchar(s[1],ord(s[0]),'');
Show(s+'');
colour := $13;
Show(copy(StrCopyright,1,i));
colour := $13;
Show('ͻ');

{ Display all the text lines. }
for i := 1 to LPS - 3 - HintLinesMax do
  begin
  colour := $13;
  Show(' ');
  colour := $1E;
  if i - 1 + RhymeTop <= RhymeLinesMax
    then Show(left(Rhyme[i-1+RhymeTop],CPL-4))
    else Show(left('',CPL-4));
  colour := $13;
  Show(' ');
  end;

ShowHintLines;
ShowLastLine;
ShowLineNumber;
end;






(*******************************************************************************
** Keyboard handling subroutines.                                             **
*******************************************************************************)

procedure DoHome;
{ Move the cursor to the begin of the line. }
begin
RhymeCursor := 1;
end;




procedure DoEnd;
{ Move the cursor to the end of the text on this line + 1. }
begin
RhymeCursor := length(Rhyme[RhymeLine]);
while (RhymeCursor > 1) and (Rhyme[RhymeLine][RhymeCursor] = ' ') do
  dec(RhymeCursor);
if (Rhyme[RhymeLine][RhymeCursor] <> ' ') and (RhymeCursor < RhymeCharsMax)
  then inc(RhymeCursor);
end;




procedure DoUp(mode : word);
{ Move the cursor one line up. }
begin
{ If we are at the first line in the file, then do nothing and exit. }
if RhymeLine = 1 then exit;

Rhyme[RhymeLine] := StripT(Rhyme[RhymeLine]);         { Strip trailing spaces. }

dec(RhymeLine);
if ((RhymeLine < RhymeTop + 3) and (RhymeTop > 1)) or (mode = 2)
  then
    begin
    dec(RhymeTop);
    ShowMenu;
    end
  else ShowLineNumber;
end;




procedure DoDown(mode : word);
begin
{ If we are at the end of the file, then show message and exit. }
if RhymeLine >= RhymeLinesMax then
  begin
  ShowErrorMessage(StrEof,true);
  exit;
  end;

Rhyme[RhymeLine] := StripT(Rhyme[RhymeLine]);         { Strip trailing spaces. }

{ Move the cursor one line down, and possibly move the top of the page. }
inc(RhymeLine);
if ((RhymeLine >= RhymeTop + LPS - 3 - HintLinesMax - 3) and
    (RhymeTop + LPS - 3 - HintLinesMax <= RhymeLinesMax)) or
   (mode = 2)
  then
    begin
    inc(RhymeTop);
    ShowMenu;
    end
  else ShowLineNumber;
end;




procedure DoLeft(mode : word);
{ Move the cursor one step left. If at the beginning of the line, then move the
  cursor to the end of the previous line. }
begin
if RhymeCursor > 1
  then dec(RhymeCursor)
  else if mode = 1 then
    begin
    DoUp(1);
    DoEnd;
    end;
end;




procedure DoRight;
{ Move the cursor one position to the right. }
begin
if RhymeCursor < RhymeCharsMax then inc(RhymeCursor);
end;




procedure DoEnter;
{ Split the line at the cursor. }
var
  i, j : word;
begin
if RhymeLine >= RhymeLinesMax then
  begin
  ShowErrorMessage(StrEof,true);
  exit;
  end;

i := RhymeLinesMax;                    { Make room by moving lower lines down. }
j := i - 1;
while j > RhymeLine do
  begin
  Rhyme[i] := Rhyme[j];
  dec(i);
  dec(j);
  end;
Rhyme[RhymeLine+1] := '';

{ If the cursor is not beyond the end of the text on the line, then split the
  line. }
if length(Rhyme[RhymeLine]) >= RhymeCursor then
  begin
  Rhyme[RhymeLine+1] := copy(Rhyme[RhymeLine],RhymeCursor,
    length(Rhyme[RhymeLine])-RhymeCursor+1);
  Delete(Rhyme[RhymeLine],RhymeCursor,
    length(Rhyme[RhymeLine])-RhymeCursor+1);
  end;

ShowMenu;                                                { Update the display. }
DoHome;                                      { Goto the beginning of the line. }
DoDown(1);                                                 { Go one line down. }
end;




procedure DoPgup;
begin
Rhyme[RhymeLine] := StripT(Rhyme[RhymeLine]);         { Strip trailing spaces. }

{ Decrement the cursor line and the top of the screen. }
if RhymeLine > LPS - 3 - HintLinesMax - 6
  then dec(RhymeLine,LPS - 3 - HintLinesMax - 6)
  else RhymeLine := 1;
if RhymeTop > LPS - 3 - HintLinesMax - 6
  then dec(RhymeTop,LPS - 3 - HintLinesMax - 6)
  else RhymeTop := 1;
ShowMenu;
end;




procedure DoPgdn;
begin
{ If at the end of the file, then show an error message and exit. }
if RhymeLine >= RhymeLinesMax then
  begin
  ShowErrorMessage(StrEof,true);
  exit;
  end;

Rhyme[RhymeLine] := StripT(Rhyme[RhymeLine]);         { Strip trailing spaces. }

{ Increment the cursor line and the top of the screen. }
if RhymeLine + LPS - 3 - HintLinesMax - 6 < RhymeLinesMax
  then inc(RhymeLine,LPS - 3 - HintLinesMax - 6)
  else RhymeLine := RhymeLinesMax;
if RhymeTop + LPS - 3 - HintLinesMax - 6 < RhymeLinesMax - LPS + 3 +
   HintLinesMax + 1
  then inc(RhymeTop,LPS - 3 - HintLinesMax - 6)
  else RhymeTop := RhymeLinesMax - LPS + 3 + HintLinesMax + 1;
ShowMenu;
end;




procedure DoDel(mode : word);
{ Delete the character under the cursor. }
var
  i : word;
begin
{ If at the end of the line, then join two lines. }
if (RhymeLine < RhymeLinesMax) and (RhymeCursor > length(Rhyme[RhymeLine]))
  then
    begin
    if mode = 2 then exit;
    while length(Rhyme[RhymeLine]) < RhymeCursor - 1 do
      Rhyme[RhymeLine] := Rhyme[RHymeLine] + ' ';
    i := RhymeCharsMax - length(Rhyme[RhymeLine]);
    if i > length(Rhyme[RhymeLine+1]) then i := length(Rhyme[RhymeLine+1]);
    if length(Rhyme[RhymeLine+1]) > 0 then
      begin
      Rhyme[RhymeLine] := Rhyme[RhymeLine] +
        copy(Rhyme[RhymeLine+1],1,i);
      delete(Rhyme[RhymeLine+1],1,i);
      end;
    if length(Rhyme[RhymeLine+1]) = 0 then
      begin
      for i := RhymeLine+1 to RhymeLinesMax-1 do
        Rhyme[i] := Rhyme[i+1];
      Rhyme[RhymeLinesMax] := '';
      end;
    ShowMenu;
    end
  else
    begin
    delete(Rhyme[RhymeLine],RhymeCursor,1);
    xy := (RhymeLine-RhymeTop+1) * CPL * 2 + 4;
    colour := $1E;
    Show(left(Rhyme[RhymeLine],CPL-4));
    end;
end;




procedure DoBackspace(mode : word);
{ Delete the character just before the cursor. }
begin
DoLeft(mode);
DoDel(mode);
end;





procedure DoHelp(which : word);
{ Show help information. }
var
  i : word;
begin
StopHintSearch('Help');

case which of
  1: for i := 1 to HelpLines1 do HintLines[i] := Help1[i];
  2: for i := 1 to HelpLines2 do HintLines[i] := Help2[i];
  3: for i := 1 to HelpLines3 do HintLines[i] := Help3[i];
  end;

case which of
  1: HintLinesMax := HelpLines1;
  2: HintLinesMax := HelpLines2;
  3: HintLinesMax := HelpLines3;
  end;
HintLinesRefresh := false;

ShowMenu;
ShowCursor(true);                                               { Show cursor. }
repeat until keypressed;
end;






procedure DoPrint;
{ Print the text. }
var
  i, j, top : word;
begin
Rhyme[RhymeLine] := StripT(Rhyme[RhymeLine]);         { Strip trailing spaces. }

{ Stop the search for hints, and show a "busy printing" message. }
StopHintSearch(StrPrinting);
ShowMenu;

top := RhymeLinesMax;
while (top >= 1) and (Rhyme[top] = '') do dec(top);
if top > 0 then
  begin
  i := 1;
  while (not keypressed) and (i <= top) do
    begin
    j := 1;
    while (not keypressed) and (j <= length(Rhyme[i])) do
      begin
      regs.AH := $05;
      regs.DL := ord(Rhyme[i][j]);
      msdos(regs);
      inc(j);
      end;
    inc(i);
    if (i > top) or keypressed then
      begin
      regs.AH := $05;
      regs.DL := $0C;
      msdos(regs);
      end;
    regs.AH := $05;
    regs.DL := $0D;
    msdos(regs);
    regs.AH := $05;
    regs.DL := $0A;
    msdos(regs);
    end;
  end;

if keypressed then                              { Forget the key just pressed. }
  begin
  regs.ah := 0;       { Fetch the key pressed into "Inkey", or wait for inkey. }
  intr($16,regs);
  end;

{ Remove the "busy printing" message. }
StopHintSearch(StrPrinting);
ShowMenu;

end;               





(*******************************************************************************
** Main loop.                                                                 **
*******************************************************************************)


function WaitInkey : word;
{ Wait for a key to be pressed, and return it as a scancode. Meanwhile, search
  for rhyming words. }
var
  HintSearchTimer : byte;
  i : word;
begin
ShowCursor(true);                                               { Show cursor. }

{ Start the search for rhyming words. If we are already searching for the
  current word, then continue that search. }
if (not keypressed) and StartHintSearch(Rhyme[RhymeLine]) then
  begin
  HintSearchTimer := mem[$0040:$006C];
  ShowMenu;        { The hintlines have been cleared, so redisplay the screen. }
  ShowCursor(true);                                             { Show cursor. }
  end;

{ Quick search the first-smaller word in the words database by iteration. }
while (not keypressed) and HintBusy and (HintInterval > 0) do HintSearchZoom;

{ Walk through the datafile testing the words for similarity with the word we
  are looking for. Stop searching at the first larger word in the file. }
while (not keypressed) and HintBusy do
  begin
  HintSearchWord;
  { If the search has finished then update the display. }
  if (not HintBusy) or (byte(mem[$0040:$006C] - HintSearchTimer) > 9) then
    begin
    HintSearchTimer := mem[$0040:$006C];
    if HintLinesRefresh then
      begin
      HintLinesRefresh := false;
      i := HintLinesMax;
      FillHintLines;
      if i <> HintLinesMax
        then ShowMenu
        else ShowHintLines;
      ShowCursor(true);
      end;
    end;
  end;

regs.ah := 0;         { Fetch the key pressed into "Inkey", or wait for inkey. }
intr($16,regs);
WaitInkey := regs.ax;

ShowCursor(false);                                         { Undisplay cursor. }
ShowErrorMessage('',false);                          { Turn error message off. }
end;






procedure MainLoop;
{ The main loop, exited only when the user presses ESC. }
var
  i, j  : word;
  Inkey : word;                                        { The last pressed key. }
  ff : text;
  Max : word;
  InputFileName : string;       { Name of the text file to store the rhyme in. }
begin
(******************************
       Show documentation.
*******************************)

{ Initialize variables. }
RhymeTop     := 1;
RhymeLine    := 1;
RhymeCursor  := 1;
HintFilePath := '';
HelpMessage := Help2[0];
HintBusy     := false;
ErrorMessage := '';
StopHintSearch('');
for i := 1 to RhymeLinesMax do Rhyme[i] := '';             { Clear text array. }

assign(ff,HelpFileName);                                 { Load the help-file. }
{$I-}
reset(ff);
if IOresult = 0
  then
    begin
    i := 1;
    while (not eof(ff)) and (i <= RhymeLinesMax) do
      begin
      readln(ff,Rhyme[i]);
      inc(i);
      end;
    close(ff);
    i := 0;
    end
  else i := 1;            { If not loadable: set flag to skip the help screen. }
{$I+}

ShowMenu;                                          { Show the complete screen. }

if i = 0 then repeat

  regs.ah := 0;       { Fetch the key pressed into "Inkey", or wait for inkey. }
  intr($16,regs);
  Inkey := regs.ax;
  ShowErrorMessage('',false);                        { Turn error message off. }

  if Inkey = $3B00 then DoHelp(2);                                       { F1. }
  if Inkey = $4100 then DoPrint;                                         { F7. }
  if Inkey = $4800 then DoUp(2);                                   { arrow-up. }
  if Inkey = $5000 then DoDown(2);                               { arrow-down. }
  if Inkey = $4900 then DoPgup;                                        { pgup. }
  if Inkey = $5100 then DoPgdn;                                        { pgdn. }
  if lo(Inkey) = 27 then exit;                                          { ESC. }

until lo(Inkey) = 13;                               { Until ENTER key pressed. }


(******************************
         Ask filename.
*******************************)

{ We ask the user for a filename using a special mode of the RHYME program. The
  text array is filled with helptext, and the user may edit line 3. He cannot
  leave this line. }

{ Initialize variables. }
RhymeTop     := 1;
RhymeLine    := 3;
HintFilePath := '';
HelpMessage := Help3[0];
HintBusy     := false;
ErrorMessage := '';
StopHintSearch('');
for i := 1 to RhymeLinesMax do Rhyme[i] := '';             { Clear text array. }
{ Load the text array with the helptext for entering a filename. }
for i := 1 to AskFilenameLines do Rhyme[i] := AskFilename[i];

{ Accept filename from the command line, or empty. }
If ParamCount = 0
  then InputFileName := ''
  else InputFileName := paramstr(1);
InputFileName := Fexpand(InputFileName);
Rhyme[3] := InputFileName;
RhymeCursor  := length(InputFileName) + 1;

ShowMenu;                                          { Show the complete screen. }

repeat

  ShowCursor(true);                                             { Show cursor. }
  regs.ah := 0;       { Fetch the key pressed into "Inkey", or wait for inkey. }
  intr($16,regs);
  Inkey := regs.ax;
  ShowCursor(false);                                             { Cursor off. }
  ShowErrorMessage('',false);                        { Turn error message off. }

  if Inkey = $5300 then DoDel(2);                                       { Del. }
  if lo(Inkey) = 8 then DoBackspace(2);                           { Backspace. }
  if Inkey = $4B00 then DoLeft(2);                               { arrow-left. }
  if Inkey = $4D00 then DoRight;                                { arrow-right. }
  if Inkey = $4700 then DoHome;                                        { home. }
  if Inkey = $4F00 then DoEnd;                                          { end. }
  if Inkey = $3B00 then DoHelp(3);                                       { F1. }

  if lo(Inkey) > 31 then                          { Process a character inkey. }
    begin
    while length(Rhyme[RhymeLine]) < RhymeCursor - 1 do
      Rhyme[RhymeLine] := Rhyme[RHymeLine] + ' ';
    colour := $1E;
    if RhymeCursor = length(Rhyme[RhymeLine]) + 1
      then
        begin
        Rhyme[RhymeLine] := Rhyme[RHymeLine] + chr(lo(Inkey));
        xy := (((RhymeLine-RhymeTop+1) * CPL) + RhymeCursor + 1) * 2;
        Show(Rhyme[RhymeLine][RhymeCursor]);
        end
      else
        begin
        Rhyme[RhymeLine] := copy(Rhyme[RhymeLine],1,RhymeCursor-1) +
          chr(lo(Inkey)) + copy(Rhyme[RhymeLine],RhymeCursor,
          length(Rhyme[RhymeLine])-RhymeCursor+1);
        xy := (RhymeLine-RhymeTop+1) * CPL * 2 + 4;
        Show(left(Rhyme[RhymeLine],CPL-4));
        end;
    if RhymeCursor < RhymeCharsMax then inc(RhymeCursor);
    end;

  { If ENTER pressed, then test if the filename is OK. }
  if lo(Inkey) = 13 then
    begin
    Rhyme[3] := Fexpand(Rhyme[3]);
    RhymeCursor := length(Rhyme[3]) + 1;
    ShowMenu;                                      { Show the complete screen. }
    i := length(Rhyme[3]);                        { Reject reserved filenames. }
    while (i > 0) and (Rhyme[3][i] <> '.') do dec(i);
    if (lower(copy(Rhyme[3],i,255)) = HintFileExt) or
       (lower(copy(Rhyme[3],i,255)) = '.pas') or
       (lower(copy(Rhyme[3],i,255)) = '.exe') or
       (lower(copy(Rhyme[3],i,255)) = '.com') then
      begin               
      ShowErrorMessage(StrReserved,true);
      Inkey := 0;
      end;
    if length(Rhyme[3]) > 0 then             { Test if the file is accessible. }
      begin
      assign(ff,Rhyme[3]);
      {$I-}
      reset(ff);
      if IOresult <> 0 then
        begin
        rewrite(ff);
        if IOresult <> 0 then
          begin
          ShowErrorMessage(StrAccess,true);
          Inkey := 0;
          end;
        end;
      close(ff);
      {$I+}
      end;
    end;

  if lo(Inkey) = 27 then exit;                          { Exit if ESC pressed. }

until lo(Inkey) = 13;                               { Until ENTER key pressed. }

InputFileName := Rhyme[3];       { Fetch the entered filename from textline 3. }


(******************************
         Start rhyming.
*******************************)

{ Initialize variables. }
RhymeCursor  := 1;
RhymeTop     := 1;
RhymeLine    := 1;
HintFilePath := '';
HelpMessage  := Help1[0];
HintBusy     := false;
ErrorMessage := '';
StopHintSearch('');
HintMessage  := StrStart;
for i := 1 to RhymeLinesMax do Rhyme[i] := '';             { Clear text array. }

ShowMenu;                                          { Show the complete screen. }

{ Determine the path where the hintfiles are. It should be the same path as the
  program. }
HintFilePath := FExpand(paramstr(0));
i := length(HintFilePath);
while (i > 0) and (HintFilePath[i] <> '\') do dec(i);
HintFilePath[0] := chr(i);


assign(ff,InputFileName);                                     { Load the file. }
{$I-}
reset(ff);
if IOresult = 0 then
  begin
  i := 1;
  while (not eof(ff)) and (i <= RhymeLinesMax) do
    begin
    readln(ff,Rhyme[i]);
    inc(i);
    end;
  close(ff);
  end;
{$I+}

ShowMenu;                                          { Show the complete screen. }

repeat

  ShowCursor(true);                                             { Show cursor. }
  Inkey := WaitInkey;

  if lo(Inkey) = 13 then DoEnter;                         { Process ENTER key. }
  if Inkey = $5300 then DoDel(1);                                       { Del. }
  if lo(Inkey) = 8 then DoBackspace(1);                           { Backspace. }
  if Inkey = $3B00 then DoHelp(1);                                       { F1. }
  if Inkey = $4100 then DoPrint;                                         { F7. }
  if Inkey = $4B00 then DoLeft(1);                               { arrow-left. }
  if Inkey = $4D00 then DoRight;                                { arrow-right. }
  if Inkey = $4700 then DoHome;                                        { home. }
  if Inkey = $4F00 then DoEnd;                                          { end. }
  if Inkey = $4800 then DoUp(1);                                   { arrow-up. }
  if Inkey = $5000 then DoDown(1);                               { arrow-down. }
  if Inkey = $4900 then DoPgup;                                        { pgup. }
  if Inkey = $5100 then DoPgdn;                                        { pgdn. }
  if Inkey = $5200 then ;                                               { Ins. }

  if lo(Inkey) > 31 then                          { Process a character inkey. }
    begin
    while length(Rhyme[RhymeLine]) < RhymeCursor - 1 do
      Rhyme[RhymeLine] := Rhyme[RHymeLine] + ' ';
    colour := $1E;
    if RhymeCursor = length(Rhyme[RhymeLine]) + 1
      then
        begin
        Rhyme[RhymeLine] := Rhyme[RHymeLine] + chr(lo(Inkey));
        xy := (((RhymeLine-RhymeTop+1) * CPL) + RhymeCursor + 1) * 2;
        Show(Rhyme[RhymeLine][RhymeCursor]);
        end
      else
        begin
        Rhyme[RhymeLine] := copy(Rhyme[RhymeLine],1,RhymeCursor-1) +
          chr(lo(Inkey)) + copy(Rhyme[RhymeLine],RhymeCursor,
          length(Rhyme[RhymeLine])-RhymeCursor+1);
        xy := (RhymeLine-RhymeTop+1) * CPL * 2 + 4;
        Show(left(Rhyme[RhymeLine],CPL-4));
        end;
    if RhymeCursor < RhymeCharsMax then inc(RhymeCursor);
    end;

until lo(Inkey) = 27;                                 { Until ESC key pressed. }

{ Write the file. }
Max := RhymeLinesMax;                                { Find the top text line. }
while (Max > 0) and (Rhyme[Max] = '') do dec(Max);
{$I-}
assign(ff,InputFileName);
if Max > 0
  then
    begin
    rewrite(ff);
    if IOresult <> 0 then reset(ff);
    if InputFileName <> '' then for i := 1 to Max do writeln(ff,Rhyme[i]);
    close(ff);
    end
  else erase(ff);
{$I+}
end;





(*******************************************************************************
** Program entry point.                                                       **
*******************************************************************************)

begin
VidStore     := nil;                                    { Initialize variable. }
HintBusy     := false;
StopHintSearch('');
ErrorMessage := '';

writeln(StrCopyright);
InitVideo;                                             { Initialise the video. }
Mainloop;                                    { The main loop, exited with ESC. }
ResetVideo;                             { Reset the original display contents. }
writeln(StrClose);
end.
